/*
 * Copyright 2017 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.call;

import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.Map;

import dk.vajhoej.record.Alignment;
import dk.vajhoej.record.ArrayField;
import dk.vajhoej.record.FieldType;
import dk.vajhoej.record.RecordException;
import dk.vajhoej.record.Struct;
import dk.vajhoej.record.StructField;
import dk.vajhoej.record.StructReader;
import dk.vajhoej.record.StructWriter;

/**
 * Class VMS contains all the glue to call native VMS code from Java using VMS calling convention.
 */
public class VMS {
	static {
		System.loadLibrary("VMSCall_shr");
		if(!setup()) {
			throw new RuntimeException("Unable to initalize VMSCall");
		}
	}
	/**
	 * Missing argument (pass of 0 by value).
	 */
	public static final Pass MISSING = (new Pass(new LongWord(0))).byValue().readOnly();
	private static String charSet = "ISO-8859-1";
	/**
	 * Get character set.
	 * @return name of character set in use
	 */
	public static String getCharSet() {
		return charSet;
	}
	/**
	 * Set character set (default is ISO-8859-1).
	 * @param charSet name of character set to use
	 */
	public static void setCharSet(String charSet) {
		VMS.charSet = charSet;
	}
	/**
	 * Class TypeWrapper.
	 * @param <T> underlying Java type
	 */
	public static class TypeWrapper<T> {
		private T v;
		public TypeWrapper(T v) {
			this.v = v;
		}
		public T getValue() {
			return v;
		}
		public void setValue(T v) {
			this.v = v;
		}
	}
	/**
	 * Class Word.
	 */
	public static class Word extends TypeWrapper<Short> {
		public Word(short v) {
			super(v);
		}
		public Word() {
			super((short)0);
		}
		public void setValue(short v) {
			setValue(new Short(v));
		}
	}
	/**
	 * Class LongWord.
	 */
	public static class LongWord extends TypeWrapper<Integer> {
		public LongWord(int v) {
			super(v);
		}
		public LongWord() {
			super(0);
		}
		public void setValue(int v) {
			setValue(new Integer(v));
		}
	}
	/**
	 * Class QuadWord.
	 */
	public static class QuadWord extends TypeWrapper<Long> {
		public QuadWord(long v) {
			super(v);
		}
		public QuadWord() {
			super(0L);
		}
		public void setValue(long v) {
			setValue(new Long(v));
		}
	}
	/**
	 * Class S Float.
	 */
	public static class SFloat extends TypeWrapper<Float> {
		public SFloat(float v) {
			super(v);
		}
		public SFloat() {
			super(0.0F);
		}
		public void setValue(float v) {
			setValue(new Float(v));
		}
	}
	/**
	 * Class T Float.
	 */
	public static class TFloat extends TypeWrapper<Double> {
		public TFloat(double v) {
			super(v);
		}
		public TFloat() {
			super(0.0);
		}
		public void setValue(double v) {
			setValue(new Double(v));
		}
	}
	/**
	 * Class CharacterString.
	 */
	public static class CharacterString extends TypeWrapper<byte[]> {
		private boolean literal;
		public CharacterString(byte[] ba) {
			super(ba);
			literal = false;
		}
		public CharacterString(String s) throws UnsupportedEncodingException {
			super(s.getBytes(charSet));
			literal = true;
		}
		public CharacterString(int n) {
			super(new byte[n]);
			literal = false;
		}
		public short getLength() {
			return (short)getValue().length;
		}
		public boolean isLiteral() {
			return literal;
		}
	}
	/**
	 * Class ByteBuffer.
	 */
	public static class ByteBuffer extends TypeWrapper<byte[]> {
		public ByteBuffer(byte[] ba) {
			super(ba);
		}
		public ByteBuffer(String s) throws UnsupportedEncodingException {
			super(s.getBytes(charSet));
		}
		public ByteBuffer(int n) {
			super(new byte[n]);
		}
		public short getLength() {
			return (short)getValue().length;
		}
	}
	private static byte[] serialize(Object obj) throws RecordException {
		StructWriter sw = new StructWriter(100000);
		sw.write(obj);
		return sw.getBytes();
	}
	private static void deserialize(byte[] b, Object obj) throws RecordException {
		StructReader sr = new StructReader(b);
		sr.read(obj);
	}
	/**
	 * Class Block.
	 */
	public static class Block<T> extends TypeWrapper<byte[]> {
		private T obj;
		public Block(T obj) throws RecordException {
			super(VMS.serialize(obj));
			this.obj = obj;
		}
		public void update() throws RecordException {
			VMS.deserialize(getValue(), obj);
		}
		public T getObject(Class<T> clz) {
			return obj;
		}
		public int getLength() {
			return getValue().length;
		}
	}
	/**
	 * Class BlockHeader.
	 */
	@Struct(alignment=Alignment.PACKED)
	public static class BlockHeader {
		@ArrayField(countprefix=4)
		@StructField(n=0,type=FieldType.INT4)
		private int[] offsets;
		public BlockHeader(int... offsets) {
			this.offsets = offsets;
		}
		public int[] getOffsets() {
			return offsets;
		}
	}
	/**
	 * Class MultiBlock.
	 */
	public static class MultiBlock extends TypeWrapper<byte[]> {
		private Object[] obj;
		private int[] ix;
		private int[] len;
		private static byte[] serialize(BlockHeader hdr, Object... obj) throws RecordException {
			byte b[][] = new byte[1 + obj.length][];
			int totlen = 0;
			b[0] = VMS.serialize(hdr);
			totlen += b[0].length;
			for(int i = 0; i < obj.length; i++) {
				b[1 + i] = VMS.serialize(obj[i]);
				totlen += b[1 + i].length;
			}
			byte[] tot = new byte[totlen];
			int ix = 0;
			for(int j = 0; j < b.length; j++) {
				System.arraycopy(b[j], 0, tot, ix, b[j].length);
				ix += b[j].length;
			}
			return tot;
		}
		private static <T2> void deserialize(byte[] all, int ix, int len, T2 obj) throws RecordException {
			byte[] b = new byte[len];
			System.arraycopy(all, ix, b, 0, len);
			VMS.deserialize(b, obj);
		}
		public MultiBlock(BlockHeader hdr, Object... obj) throws RecordException {
			super(serialize(hdr, obj));
			this.obj = obj;
			ix = new int[obj.length];
			len = new int[obj.length];
			for(int i = 0; i < obj.length; i++) {
				ix[i] = (i == 0) ? VMS.serialize(hdr).length : ix[i - 1] + len[i - 1];
				len[i] = VMS.serialize(obj[i]).length;
			}
		}
		public void update() throws RecordException {
			for(int i = 0; i < obj.length; i++) {
				deserialize(getValue(), ix[i], len[i], obj[i]);
			}
		}
		@SuppressWarnings("unchecked")
		public <T> T getObject(int blkix, Class<T> clz) throws RecordException {
			return (T)obj[blkix];
		}
		public int getLength() {
			return getValue().length;
		}
	}
	private static final int ITEM_LONGWORD_IN = 1;
	private static final int ITEM_LONGWORD_OUT = 2;
	private static final int ITEM_QUADWORD_IN = 3;
	private static final int ITEM_QUADWORD_OUT = 4;
	private static final int ITEM_CHARACTERSTRING_IN = 5;
	private static final int ITEM_CHARACTERSTRING_OUT = 6;
	/**
	 * Class Item32 representing 32 bit item.
	 */
	public static abstract class Item32 {
		private int buflen;
		private int code;
		private Object bufadr;
		private Object retlenadrorflag;
		public Item32(int someint, int code, Object bufadr, Object retlenadrorflag) {
			super();
			this.buflen = someint;
			this.code = code;
			this.bufadr = bufadr;
			this.retlenadrorflag = retlenadrorflag;
		}
		protected int getBuflen() {
			return buflen;
		}
		protected int getCode() {
			return code;
		}
		public Object getBufadr() {
			return bufadr;
		}
		protected Object getRetlenadrorflag() {
			return retlenadrorflag;
		}
		public abstract int getType();
	}
	/**
	 * Class Item32LongWordIn representing an item with an IN LongWord. 
	 */
	public static class Item32LongWordIn extends Item32 {
		public Item32LongWordIn(int code, LongWord value) {
			super(4, code, value, 0);
		}
		public Item32LongWordIn(int code, LongWord value, int flag) {
			super(4, code, value, flag);
		}
		public int getType() {
			return ITEM_LONGWORD_IN;
		}
	}
	/**
	 * Class Item32LongWordOut representing an item with an OUT LongWord. 
	 */
	public static class Item32LongWordOut extends Item32 {
		public Item32LongWordOut(int code) {
			super(4, code, new LongWord(), 0);
		}
		public LongWord getValue() {
			return (LongWord)getBufadr();
		}
		public int getType() {
			return ITEM_LONGWORD_OUT;
		}
	}
	/**
	 * Class Item32QuadWordIn representing an item with an IN QuadWord. 
	 */
	public static class Item32QuadWordIn extends Item32 {
		public Item32QuadWordIn(int code, QuadWord value) {
			super(4, code, value, 0);
		}
		public Item32QuadWordIn(int code, QuadWord value, int flag) {
			super(4, code, value, flag);
		}
		public int getType() {
			return ITEM_QUADWORD_IN;
		}
	}
	/**
	 * Class Item32QuadWordOut representing an item with an OUT QuadWord. 
	 */
	public static class Item32QuadWordOut extends Item32 {
		public Item32QuadWordOut(int code) {
			super(4, code, new QuadWord(), 0);
		}
		public QuadWord getValue() {
			return (QuadWord)getBufadr();
		}
		public int getType() {
			return ITEM_QUADWORD_OUT;
		}
	}
	/**
	 * Class Item32CharacterStringIn representing an item with an IN CharacterString. 
	 */
	public static class Item32CharacterStringIn extends Item32 {
		public Item32CharacterStringIn(int code, CharacterString value) {
			super(value.getLength(), code, value, 0);
		}
		public Item32CharacterStringIn(int code, CharacterString value, int flag) {
			super(value.getLength(), code, value, flag);
		}
		public int getType() {
			return ITEM_CHARACTERSTRING_IN;
		}
	}
	/**
	 * Class Item32CharacterStringOut representing an item with an OUT CharacterString. 
	 */
	public static class Item32CharacterStringOut extends Item32 {
		public Item32CharacterStringOut(int code, int buflen) {
			super(buflen, code, new CharacterString(buflen), new LongWord());
		}
		public CharacterString getValue() {
			return (CharacterString)getBufadr();
		}
		public short getLength() {
			return (short)(int)((LongWord)getRetlenadrorflag()).getValue();
		}
		public int getType() {
			return ITEM_CHARACTERSTRING_OUT;
		}
	}
	/**
	 * Class ItemList32 representing 32 bit item list.
	 */
	public static class ItemList32 extends TypeWrapper<Item32[]> {
		public ItemList32(Item32[] itmlst) {
			super(itmlst);
		}
	}
	public static final int ROUTINE_4ARGS_1STRETDATA = 0;
	/**
	 * Class Routine representing a routine call.
	 */
	public static class Routine extends TypeWrapper<Integer> {
		private CharacterString data;
		public Routine(int routyp) {
			super(routyp);
			data = new CharacterString(200000);
		}
		public CharacterString getData() {
			return data;
		}
	}
	private static final int BY_VALUE = 1;
	private static final int BY_REFERENCE = 2;
	private static final int BY_DESCRIPTOR = 3;
	private static final int BY_UNKNOWN = 4;
	private static final String[] passdesc = { "ByValue", "ByReference", "ByDecsriptor", "Unknown" };
	private static final int READONLY = 1;
	private static final int WRITEONLY = 2;
	private static final int READWRITE = 3;
	private static final int ACCESS_UNKNOWN = 4;
	private static final String[] accdesc = { "ReadOnly", "WriteOnly", "ReadWrite", "Unknown" };
	/**
	 * Class Pass representing passing mechanism.
	 */
	public static class Pass {
		private TypeWrapper<?> v;
		private int by;
		private int access;
		public Pass(TypeWrapper<?> v) {
			this.v = v;
			this.by = BY_UNKNOWN;
			this.access = ACCESS_UNKNOWN;
			if(v instanceof CharacterString) {
				if(((CharacterString)v).isLiteral()) {
					this.access = READONLY;
				}
			}
		}
		public TypeWrapper<?> getValue() {
			return v;
		}
		int getBy() {
			return by;
		}
		int getAccess() {
			return access;
		}
		/**
		 * Pass by value.
		 * @return this
		 */
		public Pass byValue() {
			this.by = BY_VALUE;
			return this;
		}
		/**
		 * Pass by reference.
		 * @return this
		 */
		public Pass byReference() {
			this.by = BY_REFERENCE;
			return this;
		}
		/**
		 * Pass by descriptor.
		 * @return this
		 */
		public Pass byDescriptor() {
			this.by = BY_DESCRIPTOR;
			return this;
		}
		/**
		 * Pass as readonly (IN).
		 * @return this
		 */
		public Pass readOnly() {
			this.access = READONLY;
			return this;
		}
		/**
		 * Pass as writeonly (OUT).
		 * @return this
		 */
		public Pass writeOnly() {
			this.access = WRITEONLY;
			return this;
		}
		/**
		 * Pass as read write (INOUT).
		 * @return this
		 */
		public Pass readWrite() {
			this.access = READWRITE;
			return this;
		}
	}
	/**
	 * Pass as default (unknown).
	 * @param v value
	 * @return pass
	 */
	public static Pass pass(TypeWrapper<?> v) {
		return new Pass(v);
	}
	/**
	 * Pass by value.
	 * @param v value
	 * @return pass
	 */
	public static Pass byValue(TypeWrapper<?> v) {
		return pass(v).byValue();
	}
	/**
	 * Pass by reference.
	 * @param v value
	 * @return pass
	 */
	public static Pass byReference(TypeWrapper<?> v) {
		return pass(v).byReference();
	}
	/**
	 * Pass by descriptor.
	 * @param v value
	 * @return pass
	 */
	public static Pass byDescriptor(TypeWrapper<?> v) {
		return pass(v).byDescriptor();
	}
	/**
	 * Pass as readonly.
	 * @param p pass
	 * @return pass
	 */
	public static Pass readOnly(Pass p) {
		return p.readOnly();
	}
	/**
	 * Pass as writeonly.
	 * @param p pass
	 * @return pass
	 */
	public static Pass writeOnly(Pass p) {
		return p.writeOnly();
	}
	/**
	 * Pass as readwrite.
	 * @param p pass
	 * @return pass
	 */
	public static Pass readWrite(Pass p) {
		return p.readWrite();
	}
	/**
	 * Create Java String from CharacterString and actual length.
	 * @param buf CharacterString
	 * @param len actual length
	 * @return Java String
	 * @throws UnsupportedEncodingException if Java String can not be encoded
	 */
	public static String createString(CharacterString buf, Word len) throws UnsupportedEncodingException {
		return new String(buf.getValue(), 0, len.getValue(), charSet);
	}
	/**
	 * Create Java String from CharacterString by trimming.
	 * @param buf CharacterString
	 * @return Java String
	 * @throws UnsupportedEncodingException if Java String can not be encoded
	 */
	public static String createTrimmedString(CharacterString buf) throws UnsupportedEncodingException {
		return new String(buf.getValue(), charSet).trim();
	}
	/**
	 * Create Java String from ByteBuffer and actual length.
	 * @param buf ByteBuffer
	 * @param len actual length
	 * @return Java String
	 * @throws UnsupportedEncodingException if Java String can not be encoded
	 */
	public static String createString(ByteBuffer buf, Word len) throws UnsupportedEncodingException {
		return new String(buf.getValue(), 0, len.getValue(), charSet);
	}
	/**
	 * Create Java String from ByteBuffer and actual length.
	 * @param buf ByteBuffer
	 * @param len actual length
	 * @return Java String
	 * @throws UnsupportedEncodingException if Java String can not be encoded
	 */
	public static String createString(ByteBuffer buf, int len) throws UnsupportedEncodingException {
		return new String(buf.getValue(), 0, len, charSet);
	}
	/**
	 * Create Java String from ByteBuffer by trimming.
	 * @param buf ByteBuffer
	 * @return Java String
	 * @throws UnsupportedEncodingException if Java String can not be encoded
	 */
	public static String createTrimmedString(ByteBuffer buf) throws UnsupportedEncodingException {
		return new String(buf.getValue(), charSet).trim();
	}
	private static native boolean setup();
	private static native long getFunction(String image, String function);
	private static native int callFunction(long function, int[] argtyp, int[] argacc, TypeWrapper<?>[] argval);
	private static final int WORD_BYVALUE = 1;
	private static final int LONGWORD_BYVALUE = 2;
	private static final int WORD_BYREFERENCE = 3;
	private static final int LONGWORD_BYREFERENCE = 4;
	private static final int QUADWORD_BYREFERENCE = 5;
	private static final int SFLOAT_BYREFERENCE = 6;
	private static final int TFLOAT_BYREFERENCE = 7;
	private static final int CHARACTERSTRING_BYDESCRIPTOR = 8;
	private static final int BYTEBUFFER_BYREFERENCE = 9;
	private static final int BLOCK_BYREFERENCE = 20;
	private static final int MULTIBLOCK_BYREFERENCE = 21;
	private static final int ITEMLIST32_BYREFERENCE = 22;
	private static final int ROUTINE_0 = 30;
	private static final int ROUTINE_1 = 31;
	private static final int ROUTINE_2 = 32;
	private static final int ROUTINE_3 = 33;
	private static final int ROUTINE_4 = 34;
	private static Map<String, Long> cache = new HashMap<String, Long>();
	private static String createKey(String image, String function) {
		return image + "#" + function;
	}
	/**
	 * Call VMS function.
	 * @param image name of shareable image
	 * @param function name of fucntion
	 * @param args arguments
	 * @return return status code
	 */
	public static int call(String image, String function, Pass... args) {
		String key = createKey(image, function);
		Long f = cache.get(key);
		if(f == null) {
			f = getFunction(image, function);
			if(f == 0) {
				throw new RuntimeException("Function " + function + " not found in image " + image);
			}
			cache.put(key,  f);
		}
		int narg = args.length;
		int[] argtyp = new int[narg];
		int[] argacc = new int[narg];
		TypeWrapper<?>[] argval = new TypeWrapper<?>[narg];
		for(int i = 0; i < narg; i++) {
			Pass arg = args[i];
			argacc[i] = arg.getAccess();
			argval[i] = arg.getValue();
			if(arg.getBy() == BY_VALUE && arg.getValue() instanceof Word) {
				argtyp[i] = WORD_BYVALUE;
			} else if(arg.getBy() == BY_VALUE && arg.getValue() instanceof LongWord) {
				argtyp[i] = LONGWORD_BYVALUE;
			} else if(arg.getBy() == BY_REFERENCE && arg.getValue() instanceof Word) {
				argtyp[i] = WORD_BYREFERENCE;
			} else if(arg.getBy() == BY_REFERENCE && arg.getValue() instanceof LongWord) {
				argtyp[i] = LONGWORD_BYREFERENCE;
			} else if(arg.getBy() == BY_REFERENCE && arg.getValue() instanceof QuadWord) {
				argtyp[i] = QUADWORD_BYREFERENCE;
			} else if(arg.getBy() == BY_REFERENCE && arg.getValue() instanceof SFloat) {
				argtyp[i] = SFLOAT_BYREFERENCE;
			} else if(arg.getBy() == BY_REFERENCE && arg.getValue() instanceof TFloat) {
				argtyp[i] = TFLOAT_BYREFERENCE;
			} else if(arg.getBy() == BY_DESCRIPTOR && arg.getValue() instanceof CharacterString) {
				argtyp[i] = CHARACTERSTRING_BYDESCRIPTOR;
			} else if(arg.getBy() == BY_REFERENCE && arg.getValue() instanceof ByteBuffer) {
				argtyp[i] = BYTEBUFFER_BYREFERENCE;
			} else if(arg.getBy() == BY_REFERENCE && arg.getValue() instanceof Block) {
				argtyp[i] = BLOCK_BYREFERENCE;
			} else if(arg.getBy() == BY_REFERENCE && arg.getValue() instanceof MultiBlock) {
				argtyp[i] = MULTIBLOCK_BYREFERENCE;
			} else if(arg.getBy() == BY_REFERENCE && arg.getValue() instanceof ItemList32) {
				argtyp[i] = ITEMLIST32_BYREFERENCE;
			} else if(arg.getBy() == BY_REFERENCE && arg.getValue() instanceof Routine) {
				Routine r = (Routine)arg.getValue();
				switch(r.getValue()) {
					case ROUTINE_4ARGS_1STRETDATA:
						argtyp[i] = ROUTINE_0;
						argval[i] = r.getData();
						break;
					default:
						throw new IllegalArgumentException("Unsupported routine type");
				}
			} else {
				throw new IllegalArgumentException("Unsupported pass and type combination: " + passdesc[arg.getBy() - 1] + " " + arg.getValue().getClass().getName());
			}
		}
		int retval = callFunction(f, argtyp, argacc, argval);
		try {
			for(int i = 0; i < narg; i++) {
				Pass arg = args[i];
			    if(arg.getBy() == BY_REFERENCE && arg.getValue() instanceof Block) {
			    	((Block<?>)arg.getValue()).update();
				} else if(arg.getBy() == BY_REFERENCE && arg.getValue() instanceof MultiBlock) {
			    	((MultiBlock)arg.getValue()).update();
				}
			}
		} catch (RecordException e) {
			throw new IllegalArgumentException("Block/MultiBlock not deserializable: " + e.getMessage());
		}
		return retval;
	}
	// severity bits
	public static final int STS$K_WARNING = 0;
	public static final int STS$K_SUCCESS = 1;
	public static final int STS$K_ERROR = 2;
	public static final int STS$K_INFO = 3;
	public static final int STS$K_SEVERE = 4;
	/**
	 * Combine return status code and IOSB status code.
	 * @param sts return status code
	 * @param iosb IOSB
	 * @return combined status
	 */
	public static int combineStatus(int sts, IOSB iosb) {
		if(isSuccess(sts)) {
			return iosb.getStatus();
		} else {
			return sts;
		}
	}
	/**
	 * Combine return status code and LKSB status code.
	 * @param sts return status code
	 * @param lksb LKSB
	 * @return combined status
	 */
	public static int combineStatus(int sts, LKSB lksb) {
		if(isSuccess(sts)) {
			return lksb.getStatus();
		} else {
			return sts;
		}
	}
	/**
	 * Combine return status code and IOSB status code.
	 * @param sts return status code
	 * @param iosb IOSB
	 * @return combined status
	 * @throws RecordException if internal error
	 */
	public static int combineStatusIOSB(int sts, Block<IOSB> iosb) throws RecordException {
		return combineStatus(sts, iosb.getObject(IOSB.class));
	}
	/**
	 * Combine return status code and IOSB status code.
	 * @param sts return status code
	 * @param iosb IOSB
	 * @return combined status
	 * @throws RecordException if internal error
	 */
	public static int combineStatusIOSBX(int sts, Block<IOSBX> iosb) throws RecordException {
		return combineStatus(sts, iosb.getObject(IOSBX.class));
	}
	/**
	 * Combine return status code and LKSB status code.
	 * @param sts return status code
	 * @param lksb LKSB
	 * @return combined status
	 * @throws RecordException if internal error
	 */
	public static int combineStatusLKSB(int sts, Block<LKSB> lksb) throws RecordException {
		return combineStatus(sts, lksb.getObject(LKSB.class));
	}
	/**
	 * Test if status is success.
	 * @param sts return status code
	 * @return success status
	 */
	public static boolean isSuccess(int sts) {
		return (sts & STS$K_SUCCESS) != 0;
	}
	/**
	 * Test if status is success.
	 * @param sts return status code
	 * @param iosb IOSB
	 * @return success status
	 */
	public static boolean isSuccess(int sts, IOSB iosb) {
		return isSuccess(combineStatus(sts, iosb));
	}
	/**
	 * Test if status is success.
	 * @param sts return status code
	 * @param lksb LKSB
	 * @return success status
	 */
	public static boolean isSuccess(int sts, LKSB lksb) {
		return isSuccess(combineStatus(sts, lksb));
	}
	/**
	 * Test if status is success.
	 * @param sts return status code
	 * @param iosb IOSB
	 * @return success status
	 * @throws RecordException if internal error
	 */
	public static boolean isSuccessIOSB(int sts, Block<IOSB> iosb) throws RecordException {
		return isSuccess(combineStatusIOSB(sts, iosb));
	}
	/**
	 * Test if status is success.
	 * @param sts return status code
	 * @param iosb IOSB
	 * @return success status
	 * @throws RecordException if internal error
	 */
	public static boolean isSuccessIOSBX(int sts, Block<IOSBX> iosb) throws RecordException {
		return isSuccess(combineStatusIOSBX(sts, iosb));
	}
	/**
	 * Test if status is success.
	 * @param sts return status code
	 * @param lksb LKSB
	 * @return success status
	 * @throws RecordException if internal error
	 */
	public static boolean isSuccessLKSB(int sts, Block<LKSB> lksb) throws RecordException {
		return isSuccess(combineStatusLKSB(sts, lksb));
	}
}
