/*
 * Copyright 2017 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.call.lib;

import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import dk.vajhoej.record.RecordException;

import static dk.vajhoej.vms.call.VMS.*; 
import static dk.vajhoej.vms.call.lib.VMSCodes.*;

/**
 * Class SYS_Util is a high level wrapper library for system services (SYS$ functions).
 */
public class SYS_Util extends SYS {
	/**
	 * Call SYS$BRKTHRUW to send message to terminals.
	 * @param msg message
	 * @param terminal terminal (*=all)
	 * @return return status code
	 * @throws RecordException if internal error
	 * @throws UnsupportedEncodingException if message or terminal can not be encoded
	 */
	public static Result<Void> SYS$BRKTHRUW_terminal(String msg, String terminal) throws UnsupportedEncodingException, RecordException {
		if(terminal.equals("*")) {
			return SYS$BRKTHRUW(msg, BRK$C_ALLTERMS);
		} else {
			return SYS$BRKTHRUW(msg, BRK$C_DEVICE, terminal);
		}
	}
	/**
	 * Call SYS$BRKTHRUW to send message to users.
	 * @param msg message
	 * @param user username (*=all)
	 * @return return status code
	 * @throws RecordException if internal error
	 * @throws UnsupportedEncodingException if message or username can not be encoded
	 */
	public static Result<Void> SYS$BRKTHRUW_user(String msg, String user) throws UnsupportedEncodingException, RecordException {
		if(user.equals("*")) {
			return SYS$BRKTHRUW(msg, BRK$C_ALLUSERS);
		} else {
			return SYS$BRKTHRUW(msg, BRK$C_USERNAME, user);
		}
	}
	/**
	 * Call SYS$GETQUIW to get all jobs and call callback for each.
	 * @param queuename queue name
	 * @param itms array of items
	 * @param proc callback
	 * @throws UnsupportedEncodingException if error encoding queue name or decoding value
	 * @throws RecordException of internal error
	 */
	public static void SYS$GETQUIW_QueueJobs(String queuename, int[] itms, JobProcessor<Map<Integer,Object>> proc) throws UnsupportedEncodingException, RecordException {
		if(!contains(itms, QUI$_ENTRY_NUMBER)) {
			throw new RuntimeException("QUI$_ENTRY_NUMBER not specified");
		}
		QUI_CTX ctx = new QUI_CTX();
		ctx.SYS$GETQUIW_Queue(queuename);
		for(;;) {
			Result<Map<Integer,Object>> temp = ctx.SYS$GETQUIW_Job(itms);
			if(!isSuccess(temp.getStatus())) break;
			List<String> filelist = new ArrayList<String>();
			for(;;) {
				Result<Map<Integer,Object>> file = ctx.SYS$GETQUIW_File(new int[] { QUI$_FILE_SPECIFICATION });
				if(!isSuccess(file.getStatus())) break;
				filelist.add((String)file.getResult().get(QUI$_FILE_SPECIFICATION));
			}
			temp.getResult().put(QUI$_FILELIST, filelist);
			proc.process((Integer)temp.getResult().get(QUI$_ENTRY_NUMBER), temp.getResult());
		}
		ctx.SYS$GETQUIW_Cancel();
	}
	/**
	 * Call SYS$GETQUIW to get list of all jobs and information about them.
	 * @param queuename queue name
	 * @param itms array of items
	 * @return Java List of Java Map with items
	 * @throws UnsupportedEncodingException if error encoding queue name or decoding value
	 * @throws RecordException of internal error
	 */
	public static List<Map<Integer,Object>> SYS$GETQUIW_QueueJobs(String queuename, int[] itms) throws UnsupportedEncodingException, RecordException {
		final List<Map<Integer,Object>> res = new ArrayList<Map<Integer,Object>>();
		SYS$GETQUIW_QueueJobs(queuename, itms, new JobProcessor<Map<Integer,Object>>() {
			public void process(int entry, Map<Integer, Object> o) {
				res.add(o);
			}
		});
		return res;
	}
	/**
	 * Call SYS$IDTOASC to get all identifiers and call callback for each.
	 * @param proc callback
	 * @throws UnsupportedEncodingException if error decoding UIC name
	 */
	public static void SYS$IDTOASC(IDProcessor proc) throws UnsupportedEncodingException {
		ID_CTX ctx = new ID_CTX();
		for(;;) {
			Result<Pair<Integer,String>> temp = ctx.SYS$IDTOASC();
			if(!isSuccess(temp.getStatus())) break;
			proc.process(temp.getResult().getItem1(), temp.getResult().getItem2());
		}
	}
	/**
	 * Call SYS$IDTOASC to get map of identifiers to identifier names.
	 * @param inclneg include negative (identifiers that are not UIC's)
	 * @return map of identifiers to identifier names
	 * @throws UnsupportedEncodingException if error decoding UIC name
	 */
	public static Map<Integer,String> SYS$IDTOASC(boolean inclneg) throws UnsupportedEncodingException {
		final boolean inclneg2 = inclneg;
		final Map<Integer,String> res = new HashMap<Integer,String>();
		SYS$IDTOASC(new IDProcessor() {
			public void process(int id, String idname) {
				if(inclneg2 || id > 0) res.put(id, idname);
			}
		});
		return res;
	}
	/**
	 * Call SYS$PROCESS_SCAN to get all PID's and call SYS$GETJPIW for each PID to get list of items and call callback for each..
	 * @param code code for SYS$PROCESS_SCAN
	 * @param value value for SYS$PROCESS_SCAN
	 * @param itms array of items
	 * @param proc callback
	 * @throws UnsupportedEncodingException if some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static void SYS$PROCESS_SCAN_and_SYS$GETJPIW(int code, String value, int[] itms, PIDProcessor<Map<Integer,Object>> proc) throws UnsupportedEncodingException, RecordException {
		if(!contains(itms, JPI$_PID)) {
			throw new RuntimeException("JPI$_PID not specified");
		}
		PID_CTX ctx = new PID_CTX();
		ctx.SYS$PROCESS_SCAN(code, value);
		for(;;) {
			Result<Map<Integer,Object>> temp = ctx.SYS$GETJPIW(itms);
			if(!isSuccess(temp.getStatus())) break;
			proc.process((int)(Integer)temp.getResult().get(JPI$_PID), temp.getResult());
		}
	}
	/**
	 * Call SYS$PROCESS_SCAN to get all PID's and call SYS$GETJPIW for each PID to get list of items.
	 * @param code code for SYS$PROCESS_SCAN
	 * @param value value for SYS$PROCESS_SCAN
	 * @param itms array of items
	 * @return Java List of Java Map with items
	 * @throws UnsupportedEncodingException if some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static List<Map<Integer,Object>> SYS$PROCESS_SCAN_and_SYS$GETJPIW(int code, String value, int[] itms) throws UnsupportedEncodingException, RecordException {
		final List<Map<Integer,Object>> res = new ArrayList<Map<Integer,Object>>();
		SYS$PROCESS_SCAN_and_SYS$GETJPIW(code, value, itms, new PIDProcessor<Map<Integer,Object>>() {
			public void process(int pid, Map<Integer,Object> o) {
				res.add(o);
			}
		});
		return res;
	}
	private static boolean contains(int[] a, int v) {
		for(int i = 0; i < a.length; i++) {
			if(a[i] == v) {
				return true;
			}
		}
		return false;
	}
}
