/*
 * Copyright 2017 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.call.script;

import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import dk.vajhoej.record.RecordException;
import dk.vajhoej.vms.call.lib.SYS_Util;

import static dk.vajhoej.vms.call.lib.VMSCodes.*;

/**
 * Class SYS_Script is a script friendly wrapper library for system services (SYS$ functions).
 */
public class SYS_Script extends SYS_Util {
	private static void convertFlag(Result<Map<Integer,Object>> res, int code, int[] numflags, String[] strflags) {
		if(res.getResult().containsKey(code)) {
			int val = (Integer)res.getResult().get(code);
			StringBuilder sb = new StringBuilder("");
			for(int i = 0; i < numflags.length; i++) {
				if((val & numflags[i]) != 0) {
					if(sb.length() > 0) {
						sb.append("+");
					}
					sb.append(strflags[i]);
				}
			}
			res.getResult().put(code, sb.toString());
		}
	}
	/**
	 * Call SYS$BRKTHRUW to send message to users/terminals.
	 * @param msg message
	 * @param typ type
	 * @return return status code
	 * @throws RecordException if internal error
	 * @throws UnsupportedEncodingException if message can not be encoded
	 */
	public static Result<Void> SYS$BRKTHRUW(String msg, String typ) throws RecordException, UnsupportedEncodingException {
		return SYS$BRKTHRUW(msg, VMSCodesNames.nameToNumber("BRK$C", typ));
	}
	/**
	 * Call SYS$BRKTHRUW to send message to users/terminals.
	 * @param msg message
	 * @param typ type
	 * @param to recipient
	 * @return return status code
	 * @throws RecordException if internal error
	 * @throws UnsupportedEncodingException if message or to can not be encoded
	 */
	public static Result<Void> SYS$BRKTHRUW(String msg, String typ, String to) throws RecordException, UnsupportedEncodingException {
		return SYS$BRKTHRUW(msg, VMSCodesNames.nameToNumber("BRK$C", typ), to);
	}
	/**
	 * Call SYS$CRELNM to create logical name.
	 * @param flag logical name attribute
	 * @param tblnam logical table name
	 * @param lognam logical name
	 * @param acmode access mode
	 * @param logval logical values
	 * @return return status code
	 * @throws UnsupportedEncodingException if logical table name or logical name or logical vale can not be encoded
	 */
	public static Result<Void> SYS$CRELNM(int flag, String tblnam, String lognam, String acmode, String[] logval) throws UnsupportedEncodingException {
		return SYS$CRELNM(flag, tblnam, lognam, VMSCodesNames.nameToNumber("PSL$C", acmode), logval);
	}
	/**
	 * Call SYS$ENQW to convert a lock.
	 * @param lckid lock ID
	 * @param mode lock mode
	 * @return return status code
	 * @throws RecordException if internal error
	 */
	public static Result<Void> SYS$ENQW(int lckid, String mode) throws RecordException {
		return SYS$ENQW(lckid, VMSCodesNames.nameToNumber("LCK$K", mode));
	}
	/**
	 * Call SYS$ENQW to take a lock.
	 * @param resnam resource name
	 * @param mode lock mode
	 * @return return status code + lock id
	 * @throws RecordException if internal error
	 * @throws UnsupportedEncodingException if resource name can not be encoded
	 */
	public static Result<Integer> SYS$ENQW(String resnam, String mode) throws RecordException, UnsupportedEncodingException {
		return SYS$ENQW(resnam, VMSCodesNames.nameToNumber("LCK$K", mode));
	}
	/**
	 * Call SYS$GETDVIW to get a list of items.
	 * @param devnam device name
	 * @param itms array of items
	 * @return return status code + Java Map with items
	 * @throws UnsupportedEncodingException if device name can not be encoded or some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<String,Object>> SYS$GETDVIW(String devnam, String[] itms) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.nameToNumber("DVI$", itms);
		Result<Map<Integer,Object>> res = SYS$GETDVIW(devnam, numitms);
		return new Result<Map<String,Object>>(res.getStatus(), VMSCodesNames.numberToName("DVI$", res.getResult()));
	}
	/**
	 * Call SYS$GETDVIW to get a list of all items.
	 * @param devnam device name
	 * @return return status code + Java Map with items
	 * @throws UnsupportedEncodingException if device name can not be encoded or some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<String,Object>> SYS$GETDVIW(String devnam) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.all("DVI$");
		Result<Map<Integer,Object>> res = SYS$GETDVIW(devnam, numitms);
		return new Result<Map<String,Object>>(res.getStatus(), VMSCodesNames.numberToName("DVI$", res.getResult()));
	}
	/**
	 * Call SYS$GETJPIW to get a list of items.
	 * @param pid PID
	 * @param itms array of items
	 * @return return status code + Java Map with items
	 * @throws UnsupportedEncodingException if some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<String,Object>> SYS$GETJPIW(int pid, String[] itms) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.nameToNumber("JPI$", itms);
		Result<Map<Integer,Object>> res = SYS$GETJPIW(pid, numitms);
		return new Result<Map<String,Object>>(res.getStatus(), VMSCodesNames.numberToName("JPI$", res.getResult()));
	}
	/**
	 * Call SYS$GETJPIW to get a list of all items.
	 * @param pid PID
	 * @return return status code + Java Map with items
	 * @throws UnsupportedEncodingException if some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<String,Object>> SYS$GETJPIW(int pid) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.all("JPI$");
		Result<Map<Integer,Object>> res = SYS$GETJPIW(pid, numitms);
		return new Result<Map<String,Object>>(res.getStatus(), VMSCodesNames.numberToName("JPI$", res.getResult()));
	}
	/**
	 * Call SYS$GETQUIW to get information about entry.
	 * @param entry entry number
	 * @param itms array of items
	 * @return status code + Java Map of items
	 * @throws UnsupportedEncodingException if some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<String,Object>> SYS$GETQUIW_Entry(int entry,String[] itms) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.nameToNumber("QUI$", itms);
		Result<Map<Integer,Object>> res = SYS$GETQUIW_Entry(entry, numitms);
		return new Result<Map<String,Object>>(res.getStatus(), VMSCodesNames.numberToName("QUI$", res.getResult()));
	}
	/**
	 * Call SYS$GETQUIW to get information about queue.
	 * @param queuename queue name
	 * @param itms array of items
	 * @return status code + Java Map of items
	 * @throws UnsupportedEncodingException if queue name can not be encoded or some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<String,Object>> SYS$GETQUIW_Queue(String queuename, String[] itms) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.nameToNumber("QUI$", itms);
		Result<Map<Integer,Object>> res = SYS$GETQUIW_Queue(queuename, numitms);
		convertFlag(res, QUI$_QUEUE_STATUS,
				    new int[] { QUI$M_QUEUE_IDLE, QUI$M_QUEUE_STALLED, QUI$M_QUEUE_STOPPED, QUI$M_QUEUE_BUSY},
				    new String[] { "IDLE", "STALLED", "STOPPED", "BUSY" });
		convertFlag(res, QUI$_JOB_STATUS,
				    new int[] { QUI$M_JOB_EXECUTING, QUI$M_JOB_PENDING, QUI$M_JOB_STALLED },
				    new String[] { "EXECUTING", "PENDING", "STALLED" });
		return new Result<Map<String,Object>>(res.getStatus(), VMSCodesNames.numberToName("QUI$", res.getResult()));
	}
	/**
	 * Call SYS$GETQUIW to get list of all jobs and information about them.
	 * @param queuename queue name
	 * @param itms array of items
	 * @return Java List of Java Map with items
	 * @throws UnsupportedEncodingException if error encoding queue name or decoding value
	 * @throws RecordException of internal error
	 */
	public static List<Map<String,Object>> SYS$GETQUIW_QueueJobs(String queuename, String[] itms) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.nameToNumber("QUI$", itms);
		List<Map<Integer,Object>> res = SYS$GETQUIW_QueueJobs(queuename, numitms);
		return VMSCodesNames.numberToName("QUI$", res);
	}
	/**
	 * Call SYS$GETSYIW to get information about system.
	 * @param nodename node name
	 * @param itms array of items
	 * @return status code + Java Map of items
	 * @throws UnsupportedEncodingException if some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<String,Object>> SYS$GETSYIW(String nodename, String[] itms) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.nameToNumber("SYI$", itms);
		Result<Map<Integer,Object>> res = SYS$GETSYIW(nodename, numitms);
		return new Result<Map<String,Object>>(res.getStatus(), VMSCodesNames.numberToName("SYI$", res.getResult()));
	}
	/**
	 * Call SYS$GETSYIW to get information about system.
	 * @param itms array of items
	 * @return status code + Java Map of items
	 * @throws UnsupportedEncodingException if some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<String,Object>> SYS$GETSYIW(String[] itms) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.nameToNumber("SYI$", itms);
		Result<Map<Integer,Object>> res = SYS$GETSYIW(numitms);
		return new Result<Map<String,Object>>(res.getStatus(), VMSCodesNames.numberToName("SYI$", res.getResult()));
	}
	/**
	 * Call SYS$GETUAI to get a list of items.
	 * @param usrnam username
	 * @param itms array of items
	 * @return return status code +  Java Map with items
	 * @throws UnsupportedEncodingException if username can not be encoded or some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<String,Object>> SYS$GETUAI(String usrnam, String[] itms) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.nameToNumber("UAI$", itms);
		Result<Map<Integer,Object>> res = SYS$GETUAI(usrnam, numitms);
		return new Result<Map<String,Object>>(res.getStatus(), VMSCodesNames.numberToName("UAI$", res.getResult()));
	}
	/**
	 * Call SYS$GETUAI to get a list of all items.
	 * @param usrnam username
	 * @return return status code +  Java Map with items
	 * @throws UnsupportedEncodingException if username can not be encoded or some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<String,Object>> SYS$GETUAI(String usrnam) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.all("UAI$");
		Result<Map<Integer,Object>> res = SYS$GETUAI(usrnam, numitms);
		return new Result<Map<String,Object>>(res.getStatus(), VMSCodesNames.numberToName("UAI$", res.getResult()));
	}
	/**
	 * Call SYS$IDTOASC to get list of all identifiers.
	 * @param inclneg include negative (identifiers that are not UIC's)
	 * @return list of all identifiers
	 * @throws UnsupportedEncodingException if some identifier names can not be decoded
	 */
	public static List<String> SYS$IDTOASC_AsList(boolean inclneg) throws UnsupportedEncodingException {
		return new ArrayList<String>(SYS$IDTOASC(inclneg).values());
	}
	/**
	 * Call SYS$PROCESS_SCAN to get all PID's and call SYS$GETJPIW for each PID to get list of items.
	 * @param code code for SYS$PROCESS_SCAN
	 * @param value value for SYS$PROCESS_SCAN
	 * @param itms array of items
	 * @return Java List of Java Map with items
	 * @throws UnsupportedEncodingException if value can not be encoded or some string values can not be encoded
	 * @throws RecordException if internal error
	 */
	public static List<Map<String,Object>> SYS$PROCESS_SCAN_and_SYS$GETJPIW_AsList(String code, String value, String[] itms) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.nameToNumber("JPI$", itms);
		List<Map<Integer,Object>> res = SYS$PROCESS_SCAN_and_SYS$GETJPIW(VMSCodesNames.nameToNumber("PSCAN$", code), value, numitms);
		return VMSCodesNames.numberToName("JPI$", res);
	}
	/**
	 * Call SYS$PROCESS_SCAN to get all PID's and call SYS$GETJPIW for each PID to get list of all items.
	 * @param code code for SYS$PROCESS_SCAN
	 * @param value value for SYS$PROCESS_SCAN
	 * @return Java List of Java Map with items
	 * @throws UnsupportedEncodingException if value can not be encoded or some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static List<Map<String,Object>> SYS$PROCESS_SCAN_and_SYS$GETJPIW_AsList(String code, String value) throws UnsupportedEncodingException, RecordException {
		int[] numitms = VMSCodesNames.all("JPI$");
		List<Map<Integer,Object>> res = SYS$PROCESS_SCAN_and_SYS$GETJPIW(VMSCodesNames.nameToNumber("PSCAN$", code), value, numitms);
		return VMSCodesNames.numberToName("JPI$", res);
	}
	/**
	 * Call SYS$TRNLNM to translate logical name.
	 * @param tblnam table name
	 * @param lognam logical name
	 * @param acmode access mode
	 * @param index logical name index
	 * @return return status code + translate logical name
	 * @throws UnsupportedEncodingException if table or logical name can not be encoded or translation can not be decoded
	 */
	public static Result<String> SYS$TRNLNM(String tblnam, String lognam, String acmode, int index) throws UnsupportedEncodingException {
		return SYS$TRNLNM(tblnam, lognam, VMSCodesNames.nameToNumber("PSL$C", acmode), index);
	}
}
